!=============================================================================!
!=============================================================================!
!                         error handling                                      !
!=============================================================================!
subroutine error(string,s)
  character (len=*), intent(in) :: string
  integer, intent(in) :: s
  integer, save :: n_warn  
  integer :: n_warn_max = 100

  if (s==1.or.s==11) then                   ! warning message
     write(*,*) 
     write(*,*) 'Warning:'
     write(*,*) string
     write(99,*) string
     if (s==1) then                         ! warning 
        n_warn = n_warn+1
        if (n_warn>n_warn_max) then
           write(*,*)
           write(*,*) 'more than',n_warn_max,' warnings!'
           write(*,*)
           write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
           write(*,*) '!  ELMAG 3.01 stops program excecution  !'
           write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
           stop
        endif
     endif
  endif
  if (s==0) then                    ! error
     write(*,*)
     write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
     write(*,*) '!   a serious error:                    !'
     write(99,*)'!   a serious error:                    !'
     write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
     write(*,*)
     write(*,*) string
     write(99,*)string
     write(*,*)
     write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
     write(*,*) '!  ELMAG 3.01 stops program excecution  !'
     write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
     stop
  endif

end subroutine error
!=============================================================================!
!=============================================================================!
!          random number generator from Numerical Recipes (Fortran90)         !
!=============================================================================!
function ran0()
  use internal, only : iseed
  implicit none
  integer, parameter :: K4B=selected_int_kind(9)
!  integer(K4B), intent(inout) :: iseed
  double precision ran0
  integer(K4B),parameter :: IA=16807,IM=2147483647,IQ=127773,IR=2836
  real, save :: am
  integer(K4B), save :: ix=-1, iy=-1,k

  if (iseed <= 0 .or. iy < 0) then
     am = nearest(1.e0,-1.e0)/IM
     iy = ior(ieor(888889999,abs(iseed)),1)
     ix = ieor(777755555,abs(iseed))
     iseed = abs(iseed)+1
  end if
  ix = ieor(ix,ishft(ix,13))
  ix = ieor(ix,ishft(ix,-17))
  ix = ieor(ix,ishft(ix,5))
  k = iy/IQ
  iy = IA*(iy-k*IQ)-IR*k
  if (iy<0) iy = iy+IM
  ran0 = am*ior(iand(IM,ieor(ix,iy)),1)
end function ran0
!=============================================================================!
!=============================================================================!
double precision function psran()
  implicit none
  double precision ran0
  psran=ran0()
end function psran
!=============================================================================!
!=============================================================================!
! Performs polynomial interpolation. The first term F(NN) contains the  values.
! of the function to interpolate, A(NN) contains the corresponding x-values,  .
! x is the point at which we want to evaluate the function and the last 
! parameter MM corresponds to the order of the polynomial 
! interpolation 1<MM<9 (set MM=3 if you don't know). This interpolating 
! routine requires real numbers in singol precision and an ordering
! AA(i)<AA(i+1) as input.
!=============================================================================!
double precision function DIVDIF(F,A,NN,X,MM)
  implicit none
  integer n,m,nn,mm,mplus,i,ip,ix,iy,isub,j,l,mid,npts,mmax
  double precision x,A(NN),F(NN),T(20),D(20),sum
  LOGICAL EXTRA
  DATA MMAX/10/
!
!  TABULAR INTERPOLATION USING SYMMETRICALLY PLACED ARGUMENT POINTS.
!
!  START.  FIND SUBSCRIPT IX OF X IN ARRAY A.
  IF( (NN.LT.2) .OR. (MM.LT.1) ) GO TO 20
  N=NN
  M=MM
  MPLUS=M+1
  IX=0
  IY=N+1
 !     (SEARCH INCREASING ARGUMENTS.)
1 MID=(IX+IY)/2
  IF(X.GE.A(MID)) GO TO 2
  IY=MID
  GO TO 3
!        (IF TRUE.)
2 IX=MID
3 IF(IY-IX.GT.1) GO TO 1
  GO TO 7
!  COPY REORDERED INTERPOLATION POINTS INTO (T(I),D(I)), SETTING
!  *EXTRA* TO TRUE IF M+2 POINTS TO BE USED.
7 NPTS=M+2-MOD(M,2)
  IP=0
  L=0
  GO TO 9
8 L=-L
  IF(L.GE.0) L=L+1
9 ISUB=IX+L
  IF((1.LE.ISUB).AND.(ISUB.LE.N)) GO TO 10
!        (SKIP POINT.)
  NPTS=MPLUS
  GO TO 11
!        (INSERT POINT.)
10 IP=IP+1
  T(IP)=A(ISUB)
  D(IP)=F(ISUB)
11 IF(IP.LT.NPTS) GO TO 8
  EXTRA=NPTS.NE.MPLUS
!
!  REPLACE D BY THE LEADING DIAGONAL OF A DIVIDED-DIFFERENCE TABLE, SUP-
!  PLEMENTED BY AN EXTRA LINE IF *EXTRA* IS TRUE.
  DO L=1,M
     IF(.NOT.EXTRA) GO TO 12
     ISUB=MPLUS-L
     D(M+2)=(D(M+2)-D(M))/(T(M+2)-T(ISUB))
12   I=MPLUS
     DO J=L,M
        ISUB=I-L
        D(I)=(D(I)-D(I-1))/(T(I)-T(ISUB))
        I=I-1
     end DO
  end DO        
!
!  EVALUATE THE NEWTON INTERPOLATION FORMULA AT X, AVERAGING TWO VALUES
!  OF LAST DIFFERENCE IF *EXTRA* IS TRUE.
  SUM=D(MPLUS)
  IF(EXTRA) SUM=0.5*(SUM+D(M+2))
  J=M
  DO L=1,M
     SUM=D(J)+(X-T(J))*SUM
     J=J-1
  end DO
  DIVDIF=SUM
  RETURN

20 IF(MM.LT.1) WRITE(*,101) MM
  IF(NN.LT.2) WRITE(*,102) NN

!            DIVDIF=999999999999999999.
  DIVDIF=999999999.e9

101 FORMAT( 7X, 'FUNCTION DIVDIF ... M =',I6,' IS LESS THAN 1')
102 FORMAT( 7X, 'FUNCTION DIVDIF ... N =',I6,' IS LESS THAN 2')

end function DIVDIF
!=============================================================================!
!=============================================================================!
!            injection time t as function of source distance r                !
!=============================================================================!
double precision function t_r(r)
  implicit none
  integer n,n1
  parameter(n=310,n1=311)
  double precision r,t_i(0:n),r_i(0:n)
  double precision divdif
  common /data_div3/ t_i,r_i
	
  t_r = DIVDIF(t_i,r_i,n1,r,3) 

end function t_r
!=============================================================================!
!=============================================================================!
!               source distance r as function of redshift z                   !
!=============================================================================!
double precision function r_z(z)
  implicit none
  integer n,n1
  parameter(n=310,n1=311)
  double precision z,z_i(0:n),r_i(0:n)
  double precision divdif
  common /data_div4/ z_i,r_i
	
  r_z = DIVDIF(r_i,z_i,n,z,3) 

end function r_z
!=============================================================================!
!=============================================================================!
!             redshift z  as function of source distance r                    !
!=============================================================================!
double precision function z_r(r)
  implicit none
  integer n,n1
  parameter(n=310,n1=311)
  double precision r,r_i(0:n),z_i(0:n)
  double precision divdif
  common /data_div4/ z_i,r_i
        
  z_r = DIVDIF(z_i,r_i,n1,r,3) 

end function z_r
!=============================================================================!
!=============================================================================!
!            age of a flat Universe as function of redshift z                 !
!=============================================================================!
double precision function t_z(z)
  implicit none
  integer n
  parameter(n=310)
  double precision t_i(0:n),z_i(0:n),xp,yp
  double precision divdif
  double precision z
  common /data_div1/ t_i,z_i
  
  xp = z
  yp = DIVDIF(t_i,z_i,n,xp,3) 
  t_z = yp           

end function t_z
!=============================================================================!
!=============================================================================!
!            redshift z of a flat Universe as function of age t/t_0           !
!=============================================================================!
double precision function z_t(t)
  implicit none
  integer n,n1
  parameter(n=310,n1=311)
  double precision t,t_i(0:n),z_i(0:n)
  double precision divdif
  common /data_div2/ t_i,z_i

  z_t = DIVDIF(z_i,t_i,n1,t,3) 

end function z_t
!=============================================================================!
!                        bilinear interpolation                               !
!=============================================================================!
!=============================================================================!
subroutine bilintp(x1a,x2a,ya,x1,x2,y)
  implicit none
  integer, parameter :: n=2
  double precision x1,x2,y,x1a(n),x2a(n),ya(n,n)
  double precision y1,y2,y3,y4,t,u

  y1 = ya(1,1)
  y2 = ya(2,1)
  y3 = ya(1,2)
  y4 = ya(2,2)

  t = (x1-x1a(1))/(x1a(2)-x1a(1))
  u = (x2-x2a(1))/(x2a(2)-x2a(1))

  y = (1.d0-t)*(1.d0-u)*y1 + t*(1.d0-u)*y2 + t*u*y3  + (1.d0-t)*u*y4 

end subroutine bilintp
!=============================================================================!
!=============================================================================!
!                       localising an array entry                             !
!=============================================================================!
subroutine hunt(xx,n,x,jlo)
  implicit none
  integer jlo,n
  double precision x,xx(n)
  integer inc,jhi,jm
  logical ascnd
      
  ascnd=xx(n).ge.xx(1)
  if (jlo.le.0.or.jlo.gt.n) then
     jlo=0
     jhi=n+1
     goto 3
  endif
  inc=1
  if (x.ge.xx(jlo).eqv.ascnd) then
1    jhi=jlo+inc
     if (jhi.gt.n) then
        jhi=n+1
     else if (x.ge.xx(jhi).eqv.ascnd) then
        jlo=jhi
        inc=inc+inc
        goto 1
     endif
  else
     jhi=jlo
2    jlo=jhi-inc
     if (jlo.lt.1) then
        jlo=0
     else if (x.lt.xx(jlo).eqv.ascnd) then
        jhi=jlo
        inc=inc+inc
        goto 2
     endif
  endif
3 if (jhi-jlo.eq.1) then
     if (x.eq.xx(n)) jlo=n-1
     if (x.eq.xx(1)) jlo=1
     return
  endif
  jm=(jhi+jlo)/2
  if (x.ge.xx(jm).eqv.ascnd) then
     jlo=jm
  else
     jhi=jm
  endif
  goto 3
end subroutine hunt
!=============================================================================!
!=============================================================================!
!      subroutines for solving ode adapted from  Numerical Recipes            !
!=============================================================================!
subroutine odeint(ystart,nvar,x1,x2,eps,h1,hmin,nok,nbad,derivs,rkqs,tf)
  implicit none
  integer nbad,nok,nvar,kmaxx,maxstep,nmax
  double precision eps,h1,hmin,x1,x2,ystart(nvar),tiny
  external derivs,rkqs
  parameter (maxstep=100000,nmax=50,kmaxx=200,tiny=1.d-30)
  integer i,kmax,kount,nstp
  double precision dxsav,h,hdid,hnext,x,xsav,dydx(nmax),xp(kmaxx),y(nmax),yp(nmax,kmaxx),yscal(nmax)
  logical tf
  common /path/ kmax,kount,dxsav,xp,yp
      
  x = x1
  h = sign(h1,x2-x1)
  nok = 0
  nbad = 0
  kount = 0
  do i=1,nvar
     y(i) = ystart(i)
  end do
  if ( kmax>0 ) xsav = x-2.d0*dxsav
  do nstp = 1,maxstep
     if (nstp.eq.maxstep) tf=.false.
     call derivs(x,y,dydx)
     do i=1,nvar
        yscal(i) = abs(y(i))+abs(h*dydx(i))+tiny
     end do
     if ( kmax> 0 ) then
        if (abs(x-xsav)>abs(dxsav)) then           
           if ( kount<kmax-1 )then
              kount = kount+1
              xp(kount) = x
              do i=1,nvar
                 yp(i,kount) = y(i)
              end do
              xsav = x
           end if
        end if
     end if
     if ((x+h-x2)*(x+h-x1)>0.d0) h=x2-x

     call rkqs(y,dydx,nvar,x,h,eps,yscal,hdid,hnext,derivs)
     if (hdid==h) then
        nok = nok+1
     else
        nbad = nbad+1
     end if
     if ((x-x2)*(x2-x1)>=0.d0) then
        do i=1,nvar
           ystart(i) = y(i)
        end do
        if ( kmax .ne. 0 ) then
           kount = kount+1
           xp(kount) = x
           do i=1,nvar
              yp(i,kount) = y(i)
           end do
        end if
        return
     end if
!!!     if (abs(hnext)<hmin) call error("stepsize smaller than min",1)
     h = hnext
  end do
  call error("too many steps in odeint",1)

end subroutine odeint
!=============================================================================!
!=============================================================================!
subroutine rkqs(y,dydx,n,x,htry,eps,yscal,hdid,hnext,derivs)
  implicit none
  integer n,nmax
  double precision eps,hdid,hnext,htry,x,dydx(n),y(n),yscal(n),htemp
  external derivs
  parameter (nmax=50)
  integer i
  double precision errmax,h,xnew,yerr(nmax),ytemp(nmax),safety,pgrow,pshrnk,errcon
  parameter(safety=0.9d0,pgrow=-0.2d0,pshrnk=-0.25d0,errcon=1.89d-4)
      
  h = htry           ! Set step sixe to initial trial value
1 call rkck(y,dydx,n,x,h,ytemp,yerr,derivs) !Take a step
  errmax = 0.d0      ! Evaluate accuracy
  do i=1,n
     errmax = max(errmax,abs(yerr(i)/yscal(i)))
  end do
  errmax = errmax/eps! Scale relative to required tolerance
  if ( errmax>1.d0 ) then !Truncation error too large
     h = safety*h*(errmax**pshrnk)
     if ( h<0.1d0*h) then
        h = 0.1d0*h
     end if
     htemp = safety*h*(errmax**pshrnk)
     h = sign(max(abs(htemp),0.2d0*abs(h)),h) !No more than factor 5
     xnew = x+h
     if ( xnew==x ) call error('stepsize underflow in rkqs',11) 
     goto 1
  else
     if ( errmax > errcon) then
        hnext = safety*h*(errmax**pgrow)
     else
        hnext = 5.d0*h
     end if
     hdid=h
     x=x+h
     do i=1,n
        y(i) = ytemp(i)
     end do
     return
  end if
  
end subroutine rkqs
!=============================================================================!
!=============================================================================!
subroutine rkck(y,dydx,n,x,h,yout,yerr,derivs)
  implicit none
  integer n,nmax
  double precision h,x,dydx(n),y(n),yerr(n),yout(n)
  external derivs
  parameter (nmax=6)
  double precision ak2(nmax),ak3(nmax),ak4(nmax),ak5(nmax),ak6(nmax),    &
    ytemp(nmax),a2,a3,a4,a5,a6,b21,b31,b32,b41,b42,b43,b51,b52,b53,b54,  &
    b61,b62,b63,b64,b65,c1,c3,c4,c6,dc1,dc3,dc4,dc5,dc6 
  parameter (a2=0.2d0,a3=0.3d0,a4=0.6d0,a5=1.d0,a6=0.875d0,              & 
          b21=0.2d0,b31=3.d0/40.d0,b32=9.d0/40.d0,b41=0.3d0,             &
          b42=-0.9d0,b43=1.2d0,b51=-11.d0/54.d0,b52=2.5d0,               &
          b53=-70.d0/27.d0,b54=35.d0/27.d0,                              &
          b61=1631.d0/55296.d0,b62=175.d0/512.d0,b63=575.d0/13824.d0,    &
          b64=44275.d0/110592.d0,b65=253.d0/4096.d0,c1=37.d0/378.d0,     &
          c3=250.d0/621.d0,c4=125.d0/594.d0,c6=512.d0/1771.d0,           &
          dc1=c1-2825.d0/27648.d0,dc3=c3-18575.d0/48384.d0,              & 
          dc4=c4-13525.d0/55296.d0,dc5=-277.d0/14336.d0,dc6=c6-0.25d0)

  ytemp = y+b21*h*dydx                             ! first step
       
  call derivs(x+a2*h,ytemp,ak2)                    ! second step
  ytemp = y + h*( b31*dydx + b32*ak2 )
        
  call derivs(x+a3*h,ytemp,ak3)                    ! third step
  ytemp = y + h*( b41*dydx + b42*ak2 + b43*ak3 )
        
  call derivs(x+a4*h,ytemp,ak4)                    ! fourth step
  ytemp = y + h* ( b51*dydx+b52*ak2 + b53*ak3 + b54*ak4  )
      
  call derivs(x+a5*h,ytemp,ak5)                    ! fifth step
  ytemp = y + h* ( b61*dydx + b62*ak2 + b63*ak3 + b64*ak4 + b65*ak5 )
      
  call derivs(x+a6*h,ytemp,ak6)                    ! sixth step
  yout = y + h * ( c1*dydx + c3*ak3 + c4*ak4 + c6*ak6 ) 
      
! error estimate:
  yerr = h * ( dc1*dydx + dc3*ak3 + dc4*ak4  + dc5*ak5 + dc6*ak6 ) 

end subroutine rkck
!=============================================================================!
!=============================================================================!
